import unittest
from vdutils.tests.test_convaddr import TestClass as TestClassConvAddr
from vdutils.tests.test_cordate import TestClass as TestClassCordate
from vdutils.tests.test_genpnu import TestClass as TestClassGenPnu
from vdutils.tests.test_vid import TestClass as TestClassVid


def suite():
    test_loader = unittest.TestLoader()
    test_suite = unittest.TestSuite()

    test_suite.addTest(test_loader.loadTestsFromTestCase(TestClassConvAddr))
    test_suite.addTest(test_loader.loadTestsFromTestCase(TestClassCordate))
    test_suite.addTest(test_loader.loadTestsFromTestCase(TestClassGenPnu))
    test_suite.addTest(test_loader.loadTestsFromTestCase(TestClassVid))

    return test_suite

def __run_test__():
    runner = unittest.TextTestRunner(verbosity=2)
    test_suite = suite()
    result = runner.run(test_suite)

# python -m unittest vdutils.tests.tests 
if __name__ == '__main__':
    runner = unittest.TextTestRunner(verbosity=2)
    test_suite = suite()
    result = runner.run(test_suite)


import re
import pandas as pd
from typing import (
    Any,
    List,
    Dict,
    Tuple,
    Union,
    Optional
)
from dataclasses import dataclass
from collections import defaultdict
from vdutils.library.data import (
    SGG_SPLIT_LIST,
    LAST_NM_REFINE_MAP,
    LOAD_PATTERN_FOR_SLICE,
    JIBUN_PATTERN_FOR_SLICE,
    ETC_LAND_PATTERN_FOR_SPLIT
)
from vdutils.library import Log
from vdutils.data import (
    __sep__,
    __index__,
    __encoding__,
    _get_folder_names
)
from vdutils.convaddr import ConvAddr
from vdutils.resources import resource_filename


@dataclass
class GenPnu():


    def __init__(
        self,
        base_dt: Optional[str] = None
    ):

        if base_dt is not None:
            if not isinstance(base_dt, str):
                raise TypeError("type of object('base_dt') must be string")

            if not base_dt.isdigit():
                raise ValueError("object('base_dt') should be a string consisting of numbers")

            if len(base_dt) != 8:
                raise ValueError("object('base_dt') should be a string consisting of exactly 8(YYYYMMDD) digits")
        else: pass

        self.sep = __sep__
        self.index: bool = __index__
        self.encoding: str = __encoding__
        self.origin_base_dt: str = base_dt
        self.base_dt: Optional[str] = base_dt
        self.bjd_current_df: pd.DataFrame() = None
        self.bjd_current_nm_cd_dic = defaultdict(list)
        self.bjd_dic: Dict[str, Dict[str, str]] = {}
        self.bjd_nm_change_dic: Dict[str, str] = {
            "시도명": "sido_nm",
            "시군구명": "sgg_nm",
            "읍면동명": "emd_nm",
            "리명": "ri_nm",
        }
        self.logger = Log('GeneratePnu').stream_handler("INFO")
        self._get_base_dt()
        self._get_file_names()
        self._prepare()
        self.base_dt_print: str = f"{self.base_dt[:4]}-{self.base_dt[4:6]}-{self.base_dt[6:8]}"
        self.base_folder_path = resource_filename(
            "vdutils", 
            "data/bjd"
        )


    def _find_latest_base_dt(
        self,
        base_dts: List[str]
    ) -> str:

        """
        입력된 날짜(YYYYMMDD)와 법정동 데이터 시점 리스트와 비교하여 입력된 날짜보다 과거 시점 중 최신 시점을 반환
        """

        for date in base_dts:
            if date < self.base_dt:
                return date

        # 입력된 날짜보다 작은 날짜가 없을 경우
        self.logger.info("입력된 날짜보다 이전 시점의 법정동 데이터가 존재하지 않습니다. 보유한 데이터중 최신 데이터를 적용합니다.")
        return base_dts[0]


    def _get_base_dt(self):

        """
        입력된 날짜(YYYYMMDD)와 법정동 데이터 시점 리스트와 비교하여 입력된 날짜보다 과거 시점 중 최신 시점을 반환 \n
        입력된 날짜(YYYYMMDD)가 없으면 데이터 시점 리스트 중 최신 시점을 반환
        """

        base_dts = _get_folder_names(library_name="vdutils", base_folder_path="data/bjd")
        base_dts = sorted(base_dts, reverse=True)
        try:
            if self.base_dt is None:
                self.base_dt = base_dts[0]
            else:
                self.base_dt = self._find_latest_base_dt(base_dts=base_dts)
        finally:
            self.logger.info(f"[GenPnu] 적용 법정동 데이터 시점: {self.base_dt}")


    def _get_file_names(self):
        self.file_name_bjd_current = resource_filename(
            "vdutils", 
            f"data/bjd/{self.base_dt}/bjd_current.txt"
        )


    def _get_bjd_current_df(
        self,
        file_name_bjd_current,
        input_encoding,
        input_index,
        input_sep
    ):
        try:
            self.bjd_current_df: pd.DataFrame = pd.read_csv(
                file_name_bjd_current,
                sep=input_sep,
                engine='python',
                encoding=input_encoding,
                dtype={
                    '과거법정동코드': str,
                    '법정동코드': str
                }
            )
        except Exception as e:
            self.logger.error(f"Failed to read {file_name_bjd_current}")
            self.logger.error(e)


    def _create_bjd_current_nm_cd_dic(self):
        try:
            for idx, row in self.bjd_current_df.iterrows():
                bjd_nms = []
                for bjd_nm in ["시도명", "시군구명", "읍면동명", "리명"]:
                    if not pd.isna(row[bjd_nm]):
                        bjd_nms.append(row[bjd_nm])
                bjd_nms = " ".join(bjd_nms)
                self.bjd_current_nm_cd_dic[bjd_nms].append(
                    {
                        "bjd_cd": row["법정동코드"], 
                        "deleted_dt": None if pd.isna(row["삭제일자"]) else row["삭제일자"]
                    }
                )
        except Exception as e:
            self.logger.error(f"Failed to create bjd_current_nm_cd_dic")
            self.logger.error(e)


    def _create_bjd_dic(self):
        try:
            for idx, row in self.bjd_current_df.iterrows():
                bjd_cd: str = row["법정동코드"]
                full_bjd_nm: str = row["법정동명"]
                created_dt: str = row["생성일자"]
                deleted_dt: str = row["삭제일자"]
                bjd_datas: Dict[str, str] = {}
                for bjd_nm in ["시도명", "시군구명", "읍면동명", "리명"]:
                    if not pd.isna(row[bjd_nm]):
                        bjd_datas[self.bjd_nm_change_dic[bjd_nm]] = row[bjd_nm]
                    else:
                        bjd_datas[self.bjd_nm_change_dic[bjd_nm]] = None
                bjd_datas["full_bjd_nm"] = full_bjd_nm
                bjd_datas["created_dt"] = created_dt
                bjd_datas["deleted_dt"] = None if pd.isna(deleted_dt) else deleted_dt
                self.bjd_dic[bjd_cd] = bjd_datas
        except Exception as e:
            self.logger.error(f"Failed to create bjd_dic")
            self.logger.error(e)


    def _import_convaddr(self):
        try:
            self.convaddr = ConvAddr(base_dt=self.origin_base_dt)
            self.bjd_smallest_list = self.convaddr.bjd_smallest_list
            self.convert_pnu_dic = dict((bjd_nm, bjd_cd) for bjd_cd, bjd_nm in self.convaddr.bjd_current_dic.items() if bjd_cd != "법정동코드")
        except Exception as e:
            self.logger.error(f"Failed to import vdutils.convaddr module")
            self.logger.error(e)


    def _prepare(self):
        self._get_bjd_current_df(
            file_name_bjd_current=self.file_name_bjd_current,
            input_encoding=self.encoding,
            input_index=self.index,
            input_sep=self.sep,
        )
        self._create_bjd_current_nm_cd_dic()
        self._create_bjd_dic()
        self._import_convaddr()


    def get_bjd_cd(
        self,
        bjd_nm: str,
    ) -> Dict[str, Any]:

        """
        입력된 문자열(한글 법정동명)의 법정동 코드를 반환

        Args:
            bjd_nm (str): The input should be a string consisting of Korean administrative district names.

        Raises:
            TypeError: If the 'bjd_nm' object is not of type string.
            ValueError: If the 'bjd_nm' object is not consist of only Korean characters and numbers.

        Returns:
            Dict[str, Any]: 
                "error": bool,
                "bjd_cd": Optional[str],
                "deleted_dt": Optional[str],
                "base_dt": str,
                "msg": str
        """

        if not isinstance(bjd_nm, str):
            raise TypeError("type of object('bjd_nm') must be string")

        # 한자 표기가 포함된 괄호도 허용하도록 정규식 수정
        # 한자 범위: \u4e00-\u9fff (CJK Unified Ideographs)
        # if not re.match("^[가-힣0-9]+$", bjd_nm):
        if not re.match("^[가-힣0-9 ()\u4e00-\u9fff]+$", bjd_nm): # NOTE 한자 표기 괄호 허용
            raise ValueError("object('bjd_nm') should consist of only Korean characters, numbers, and Chinese characters in parentheses")

        not_a_valid_district_response: Dict[str, Any] = {
            "error": True,
            "bjd_cd": None, 
            "deleted_dt": None,
            "base_dt": self.base_dt_print,
            "msg": f"'{bjd_nm}' is not a valid legal district name"
        }
        try:
            bjd_nm = " ".join(bjd_nm.split())

            if bjd_nm in self.bjd_current_nm_cd_dic:
                bjd_cd_list = self.bjd_current_nm_cd_dic[bjd_nm]
                if len(bjd_cd_list) > 1:
                    bjd_cd = list(
                        filter(
                            lambda bjd_cd_data: not bjd_cd_data["deleted_dt"], bjd_cd_list
                        )
                    )[0]
                else:
                    bjd_cd = bjd_cd_list[0]

                return {
                    "error": False,
                    **bjd_cd, 
                    "base_dt": self.base_dt_print,
                    "msg": ""
                }

            else:
                if len(bjd_nm.split()) == 1:
                    return not_a_valid_district_response
                else:
                    sgg = bjd_nm.split()[1]
                    if sgg in SGG_SPLIT_LIST:
                        sgg_split_nm = f"{sgg[:2]}시 {sgg[2:]}"
                        bjd_nm = bjd_nm.replace(sgg, sgg_split_nm)
                        return self.get_bjd_cd(bjd_nm)

                    last_nm = bjd_nm.split()[-1]
                    if last_nm in LAST_NM_REFINE_MAP:
                        bjd_nm = bjd_nm.replace(last_nm, LAST_NM_REFINE_MAP[last_nm])
                        return self.get_bjd_cd(bjd_nm)

            return not_a_valid_district_response

        except:
            return not_a_valid_district_response


    def get_bjd_data(
        self,
        bjd_cd: str
    ) -> Dict[str, Any]:

        """
        입력된 문자열(숫자 10자리의 법정동코드)의 법정동 데이터(각 단위 법정동명, 생성일자, 삭제일자)를 반환

        Args:
            bjd_cd (str): The Korean district code string consisting of exactly 10 digits.

        Raises:
            TypeError: If the 'bjd_cd' object is not of type string.
            ValueError: If the 'bjd_cd' object does not consist of digits only.
            ValueError: If the 'bjd_cd' object does not consist of exactly 10 digits.

        Returns:
            Dict[str, Any]: {
                "error": bool,
                "sido_nm": Optional[str],
                "sgg_nm": Optional[str],
                "emd_nm": Optional[str],
                "ri_nm": Optional[str],
                "full_bjd_nm": Optional[str],
                "created_dt": Optional[str],
                "deleted_dt": Optional[str],
                "base_dt": str,
                "msg": str
            }
        """

        if not isinstance(bjd_cd, str):
            raise TypeError("type of object('bjd_cd') must be string")

        if not bjd_cd.isdigit():
            raise ValueError("object('bjd_cd') should be a string consisting of numbers")

        if len(bjd_cd) != 10:
            raise ValueError("object('bjd_cd') should be a string consisting of exactly 10 digits")

        not_a_valid_district_response: Dict[str, Any] = {
            "error": True,
            "sido_nm": None,
            "sgg_nm": None,
            "emd_nm": None,
            "ri_nm": None,
            "full_bjd_nm": None,
            "created_dt": None,
            "deleted_dt": None,
            "base_dt": self.base_dt_print,
            "msg": f"'{bjd_cd}' is not a valid legal district code"
        }
        try:
            if bjd_cd in self.bjd_dic:
                return {"error": False, **self.bjd_dic[bjd_cd], "base_dt": self.base_dt_print, "msg": ""}
            else:
                return not_a_valid_district_response
        except Exception as e:
            return {**not_a_valid_district_response, "msg": str(e)}


    @staticmethod
    def _validate_jibun(
        jibun: Optional[str]
    ) -> bool:

        """
        입력된 지번 문자열이 올바른 형식인지 정규식을 이용하여 검증하여 반환 \n
        단, 블록지번를 의미하는 음절이 포함되거나 '*' 가 포함될 경우 예외 적용하여 True 를 리턴

        Args:
            jibun (str): Validates the format of the given address.
                The address should include '산' and only contain digits except for '산' and '-'.
                The main and sub numbers should be separated by a hyphen, and both can have a maximum of 4 digits.
                Examples: 
                    With mountain and sub-number: 산 0000-0000
                    With mountain and no sub-number: 산 0000
                    Without mountain and with sub-number: 0000-0000
                    Without mountain and without sub-number: 0000

        Raises:
            ValueError: If the 'jibun' object is not of the specified format.

        Returns:
            dict: Validation result containing 'error' (bool) and 'msg' (str).
        """

        msg = "Invalid 'jibun' format. Please follow the specified format."
        """
        Invalid 'jibun' format. Please follow the specified format.

        The address should include '산' and only contain digits except for '산' and '-'.
        The main and sub numbers should be separated by a hyphen, and both can have a maximum of 4 digits.
        Examples:
            With mountain and sub-number: 산 0000-0000
            With mountain and no sub-number: 산 0000
            Without mountain and with sub-number: 0000-0000
            Without mountain and without sub-number: 0000
        """

        if pd.isna(jibun) \
        or jibun == ""  \
        or jibun is None  \
        or jibun[0] in ["B", "가", "지"]:
            return {"error": True, "msg": msg}
        if "*" in jibun:
            return {"error": True, "msg": msg}

        jibun = jibun.replace(" ", "")
        pattern = re.compile(r'^(산\s*)?\d{1,4}-\d{1,4}$|^(산\s*)?\d{1,4}$|^\d{1,4}-\d{1,4}$|^\d{1,4}$')

        if not bool(pattern.match(jibun)):
            return {"error": True, "msg": msg}

        return {"error": False, "msg": ""}


    @staticmethod
    def _get_mountain_cd(
        jibun: str
    ) -> Tuple[str, str]:

        """
        입력된 지번 문자열(지번 문자열 적합성 확인된 입력값)에서 '산' 여부 판단하여 산코드를 반환
        """

        if jibun[0] in ["산"]:
            mountain_cd = "2"
            jibun = jibun.replace("산", "")

        else:
            mountain_cd = "1"

        return jibun, mountain_cd


    @staticmethod
    def _get_jibun_datas(
        jibun: str
    ) -> Tuple[str, str, str]:

        """
        입력된 지번 문자열(지번 문자열 적합성 확인된 입력값)에서 본번과 부번을 분리하여 번, 지 코드를 반환
        """

        jibun_split = jibun.split("-")

        if len(jibun_split) == 2:
            bun, ji = [int(num) for num in jibun_split]
            bunji_cd = "%04d%04d" % (bun, ji)
            bun = str(bun)
            ji = str(ji)

        elif len(jibun_split) == 1:
            bun = int(jibun)
            bunji_cd = "%04d0000" % (bun)
            bun = str(bun)
            ji = "0"

        else:
            bunji_cd = "00000000"
            bun, ji = "0", "0"

        return bunji_cd, bun, ji


    @staticmethod
    def _validate_pnu(
        pnu: Optional[str]
    ) -> dict:
        """
        입력된 PNU 문자열이 올바른 형식인지 정규식을 이용하여 검증하여 반환

        Args:
            pnu (str): Validates the format of the given PNU.
                The PNU must be a 19-character string consisting only of digits.

        Raises:
            ValueError: If the 'pnu' object is not of the specified format.

        Returns:
            dict: Validation result containing 'error' (bool) and 'msg' (str).
        """

        msg = "Invalid 'pnu' format. Please follow the specified format."
        """
        Invalid 'pnu' format. Please follow the specified format.

        The PNU must be a 19-character string consisting only of digits.
        Example:
            1234567890123456789
        """

        if not pnu or not isinstance(pnu, str):
            return {"error": True, "msg": msg}

        pattern = re.compile(r'^\d{19}$')
        if not pattern.match(pnu):
            return {"error": True, "msg": msg}

        return {"error": False, "msg": ""}


    def _create_generate_pnu_result(
        self,
        error=True, 
        pnu=None,
        bjd_cd=None,
        mountain_cd=None,
        bunji_cd=None, 
        bjd_datas=None,
        bun=None,
        ji=None,
        msg=""
    ) -> Dict[str, Any]:

        """
        결과 딕셔너리를 생성하는 유틸리티 함수

        Returns:
            Dict[str, Any]: {
                "error": bool,
                "pnu": Optional[str],
                "bjd_cd": Optional[str],
                "mountain_cd": Optional[str],
                "bunji_cd": Optional[str],
                "bjd_datas": {
                    "error": bool,
                    "sido_nm": Optional[str],
                    "sgg_nm": Optional[str],
                    "emd_nm": Optional[str],
                    "ri_nm": Optional[str],
                    "full_bjd_nm": Optional[str],
                    "created_dt": Optional[str],
                    "deleted_dt": Optional[str],
                    "base_dt": str,
                    "msg": str,
                },
                "bun": Optional[str],
                "ji": Optional[str],
                "base_dt": str,
                "msg": str
            }
        """

        return {
            "error": error,
            "pnu": pnu,
            "bjd_cd": bjd_cd,
            "mountain_cd": mountain_cd,
            "bunji_cd": bunji_cd,
            "bjd_datas": bjd_datas or {
                "error": True,
                "sido_nm": None,
                "sgg_nm": None,
                "emd_nm": None,
                "ri_nm": None,
                "full_bjd_nm": None,
                "created_dt": None,
                "deleted_dt": None,
                "base_dt": self.base_dt_print,
                "msg": ""
            },
            "bun": bun,
            "ji": ji,
            "base_dt": self.base_dt_print,
            "msg": msg
        }


    def generate_pnu(
        self,
        bjd_cd: str,
        jibun: str # '산'을 포함한 지번
    ) -> Dict[str, Any]:

        """
        입력된 문자열(법정동 코드, 지번)을 필지관리번호(pnu)로 변환하여 반환

        Args:
            bjd_cd (str): The Korean district code string consisting of exactly 10 digits.
            jibun (str): Validates the format of the given address.
                The address should include '산' and only contain digits except for '산' and '-'.
                The main and sub numbers should be separated by a hyphen, and both can have a maximum of 4 digits.
                Examples:
                    With mountain and sub-number: 산 0000-0000
                    With mountain and no sub-number: 산 0000
                    Without mountain and with sub-number: 0000-0000
                    Without mountain and without sub-number: 0000

        Raises:
            TypeError: If the 'bjd_cd' object is not of type string.
            TypeError: If the 'jibun' object is not of type string.
            ValueError: If the 'bjd_cd' object does not consist of digits only.
            ValueError: If the 'bjd_cd' object does not consist of exactly 10 digits.

        Returns:
            Dict[str, Any]: {
                "error": bool,
                "pnu": Optional[str],
                "bjd_cd": Optional[str],
                "mountain_cd": Optional[str],
                "bunji_cd": Optional[str],
                "bjd_datas": {
                    "error": bool,
                    "sido_nm": Optional[str],
                    "sgg_nm": Optional[str],
                    "emd_nm": Optional[str],
                    "ri_nm": Optional[str],
                    "full_bjd_nm": Optional[str],
                    "created_dt": Optional[str],
                    "deleted_dt": Optional[str],
                    "base_dt": str,
                    "msg": str,
                },
                "bun": Optional[str],
                "ji": Optional[str],
                "base_dt": str,
                "msg": str
            }
        """

        if not isinstance(bjd_cd, str):
            raise TypeError("type of object('bjd_cd') must be string")

        if not isinstance(jibun, str):
            raise TypeError("type of object('jibun') must be string")

        if not bjd_cd.isdigit():
            raise ValueError("object('bjd_cd') should be a string consisting of numbers")

        if len(bjd_cd) != 10:
            raise ValueError("object('bjd_cd') should be a string consisting of exactly 10 digits")

        is_valid_jibun = self._validate_jibun(jibun)
        if is_valid_jibun.get("error") is True:
            return self._create_generate_pnu_result(
                error=True,
                msg=is_valid_jibun.get("msg")
            )

        try:
            bjd_datas = self.get_bjd_data(bjd_cd)
            if bjd_datas.get("error") is True:
                return self._create_generate_pnu_result(
                    error=True,
                    bjd_datas=bjd_datas,
                    msg=bjd_datas.get("msg")
                )

            else:
                jibun = jibun.replace(" ", "")
                jibun, mountain_cd = self._get_mountain_cd(jibun)
                bunji_cd, bun, ji = self._get_jibun_datas(jibun)
                is_valid_pnu = self._validate_pnu(f"{bjd_cd}{mountain_cd}{bunji_cd}")

                if is_valid_pnu.get("error") is True:
                    return self._create_generate_pnu_result(
                        error=True,
                        pnu=f"{bjd_cd}{mountain_cd}{bunji_cd}",
                        bjd_cd=bjd_cd,
                        mountain_cd=mountain_cd,
                        bunji_cd=bunji_cd,
                        bjd_datas=bjd_datas,
                        bun=bun,
                        ji=ji,
                        msg=is_valid_pnu.get("msg")
                    )

                else:
                    return self._create_generate_pnu_result(
                        error=False,
                        pnu=f"{bjd_cd}{mountain_cd}{bunji_cd}",
                        bjd_cd=bjd_cd,
                        mountain_cd=mountain_cd,
                        bunji_cd=bunji_cd,
                        bjd_datas=bjd_datas,
                        bun=bun,
                        ji=ji,
                        msg=""
                    )

        except Exception as e:
            self._create_generate_pnu_result(
                error=True,
                msg=str(e)
            )


    def generate_pnu_from_bjd_nm(
        self,
        bjd_nm: str,
        jibun: str # '산'을 포함한 지번
    ) -> Dict[str, Any]:

        """
        입력된 문자열(법정동명, 지번)을 필지관리번호(pnu)로 변환하여 반환

        Args:
            bjd_nm (str): The input should be a string consisting of Korean administrative district names.
            jibun (str): Validates the format of the given address.
                The address should include '산' and only contain digits except for '산' and '-'.
                The main and sub numbers should be separated by a hyphen, and both can have a maximum of 4 digits.
                Examples:
                    With mountain and sub-number: 산 0000-0000
                    With mountain and no sub-number: 산 0000
                    Without mountain and with sub-number: 0000-0000
                    Without mountain and without sub-number: 0000

        Raises:
            TypeError: If the 'bjd_nm' object is not of type string.
            TypeError: If the 'jibun' object is not of type string.
            ValueError: If the 'bjd_nm' object is not consist of only Korean characters and numbers.

        Returns:
            Dict[str, Any]: {
                "error": bool,
                "pnu": Optional[str],
                "bjd_cd": Optional[str],
                "mountain_cd": Optional[str],
                "bunji_cd": Optional[str],
                "bjd_datas": {
                    "error": bool,
                    "sido_nm": Optional[str],
                    "sgg_nm": Optional[str],
                    "emd_nm": Optional[str],
                    "ri_nm": Optional[str],
                    "full_bjd_nm": Optional[str],
                    "created_dt": Optional[str],
                    "deleted_dt": Optional[str],
                    "base_dt": str,
                    "msg": str,
                },
                "bun": Optional[str],
                "ji": Optional[str],
                "base_dt": str,
                "msg": str
            }
        """

        if not isinstance(bjd_nm, str):
            raise TypeError("type of object('bjd_nm') must be string")

        if not isinstance(jibun, str):
            raise TypeError("type of object('jibun') must be string")

        # 한자 표기가 포함된 괄호도 허용하도록 정규식 수정
        # 한자 범위: \u4e00-\u9fff (CJK Unified Ideographs)
        # if not re.match("^[가-힣0-9]+$", bjd_nm):
        if not re.match("^[가-힣0-9 ()\u4e00-\u9fff]+$", bjd_nm):
            raise ValueError("object('bjd_nm') should consist of only Korean characters, numbers, and Chinese characters in parentheses")

        try:
            res = self.get_bjd_cd(bjd_nm=bjd_nm)
            if res.get("error") is True:
                return self._create_generate_pnu_result(
                    error=res.get("error"),
                    msg=res.get("msg")
                )
            else:
                return self.generate_pnu(
                    bjd_cd=res.get("bjd_cd"),
                    jibun=jibun
                )
        except Exception as e:
            return self._create_generate_pnu_result(
                error=True,
                msg=str(e)
            )


    def _clean_bracket_and_content(
        self,
        string: str
    ) -> str:

        """
        주소문자열의 ([]) 와 같은 괄호 및 괄호안 문자 제거하는 기능
        """

        pattern = r'[\(\[].*?[\)\]]'
        return re.sub(pattern, '', string).rstrip()


    def _split_etc_main_address(
        self,
        main_address: str,
        detail_address: str
    ) -> Tuple[Optional[str], Optional[str]]:

        """
        주소문자열에서 주(main) 주소, 주 외(main etc) 주소, 상세(detail) 주소를 추출하여 반환
        """

        not_a_valid_district_response: Dict[str, Any] = {
            "error": True,
            "main_address": None, 
            "etc_main_address": None,
            "detail_address": None,
            "base_dt": self.base_dt_print,
            "msg": f"'{detail_address}' is failed to split main address and main etc address"
        }
        try:
            if re.sub(ETC_LAND_PATTERN_FOR_SPLIT, "", detail_address):
                clean_detail_address = re.sub(ETC_LAND_PATTERN_FOR_SPLIT, "", detail_address).strip()
                etc_main_address = detail_address.replace(clean_detail_address, "")
                return {
                    "error": False,
                    "main_address": main_address,
                    "etc_main_address": etc_main_address,
                    "detail_address": clean_detail_address,
                    "base_dt": self.base_dt_print,
                    "msg": ""
                }
            else:
                return {
                    "error": False,
                    "main_address": main_address,
                    "etc_main_address": None,
                    "detail_address": detail_address,
                    "base_dt": self.base_dt_print,
                    "msg": ""
                }
        except Exception as e:
            return {**not_a_valid_district_response, "msg": str(e)}


    def _split_main_and_detail_address(
        self,
        address: str
    ) -> Dict[str, Any]:

        """
        주소문자열에서 주(main) 주소, 주 외(main etc) 주소, 상세(detail) 주소를 추출하여 반환
        """

        not_a_valid_district_response: Dict[str, Any] = {
            "error": True,
            "main_address": None, 
            "etc_main_address": None,
            "detail_address": None,
            "base_dt": self.base_dt_print,
            "msg": f"'{address}' is failed to split main address and detail address"
        }
        try:
            cleaned_address = self._clean_bracket_and_content(string=address)
            match = re.search(LOAD_PATTERN_FOR_SLICE, cleaned_address)
            if match:
                return {**not_a_valid_district_response, "msg": f"'{address}' is load address"}

            match = re.search(JIBUN_PATTERN_FOR_SLICE, cleaned_address)
            if match:
                detail_address = address.split(match.group(0))[1].strip()
                main_address = address.replace(detail_address, "").strip()
                return self._split_etc_main_address(main_address=main_address, detail_address=detail_address)
            else:
                return not_a_valid_district_response
        except Exception as e:
            return {**not_a_valid_district_response, "msg": str(e)}


    def _is_not_in_smallest_bjd(
        self,
        rest_main_address: str
    ) -> bool:

        """
        입력된 지번 문자열(법정동명이 제외된 지번 문자열)에서 최소단위 법정동이 포함되어 있는지 확인하여 반환
        """

        try:
            check_word = rest_main_address.split()[0]
        except:
            check_word = rest_main_address
        for bjd in self.bjd_smallest_list:
            if check_word == bjd:
                return False
        return True


    def _extract_bjd_from_address(
        self,
        main_address: str
    ) -> Union[str, None]:

        """
        입력된 지번 문자열에서 번지 상세주소가 제외된 지번의 법정동 문자열을 추출하여 반환
        법정동 딕셔너리에서 입력된 지번 문자열내에 존재하는 법정동이 있는지 조회하고 존재할 경우,
        입력된 지번 문자열에서 매칭된 법정동을 잘라낸 나머지 문자열에 최소 단위 법정동명이 존재하는지 확인해서 최소 단위 법정동 레벨인지를 확인
        매칭된 법정동이 최소 단위 법정동이면 반환

        Args:
            addr (str): The input korean address string.

        Raises:
            TypeError: If the 'address' object is not of type string.

        Returns:
            Union[str, None]
        """

        if not isinstance(main_address, str):
            raise TypeError("type of object('address') must be string")

        try:
            for key in self.convert_pnu_dic.keys():
                if key in main_address and self.convert_pnu_dic[key][5:] != '00000':
                    rest_main_address = main_address.replace(key, '')
                    if main_address.split(key)[1][0] == ' ' \
                    and self._is_not_in_smallest_bjd(rest_main_address=rest_main_address):
                        return key
            return None

        except Exception as e:
            return None


    def generate_pnu_from_address(
        self,
        address: str,
    ) -> Dict[str, Any]:

        """
        입력된 문자열(주소)에서 법정동, 지번을 분리 필지관리번호(pnu)로 변환하여 반환

        Args:
            bjd_nm (str): The input should be a string consisting of Korean administrative district names.
            jibun (str): Validates the format of the given address.
                The address should include '산' and only contain digits except for '산' and '-'.
                The main and sub numbers should be separated by a hyphen, and both can have a maximum of 4 digits.
                Examples:
                    With mountain and sub-number: 산 0000-0000
                    With mountain and no sub-number: 산 0000
                    Without mountain and with sub-number: 0000-0000
                    Without mountain and without sub-number: 0000

        Raises:
            TypeError: If the 'bjd_nm' object is not of type string.
            TypeError: If the 'jibun' object is not of type string.
            ValueError: If the 'bjd_nm' object is not consist of only Korean characters and numbers.

        Returns:
            Dict[str, Any]: {
                "error": bool,
                "pnu": Optional[str],
                "bjd_cd": Optional[str],
                "mountain_cd": Optional[str],
                "bunji_cd": Optional[str],
                "bjd_datas": {
                    "error": bool,
                    "sido_nm": Optional[str],
                    "sgg_nm": Optional[str],
                    "emd_nm": Optional[str],
                    "ri_nm": Optional[str],
                    "full_bjd_nm": Optional[str],
                    "created_dt": Optional[str],
                    "deleted_dt": Optional[str],
                    "base_dt": str,
                    "msg": str,
                },
                "bun": Optional[str],
                "ji": Optional[str],
                "base_dt": str,
                "msg": str
            }
        """

        if not isinstance(address, str):
            raise TypeError("type of object('address') must be string")

        try:
            is_split_address = self._split_main_and_detail_address(address=address)
            if is_split_address.get("error") is True:
                return self._create_generate_pnu_result(
                    error=is_split_address.get("error"),
                    msg=is_split_address.get("msg")
                )

            else:
                main_address = is_split_address.get("main_address")
                bjd_nm = self._extract_bjd_from_address(main_address=main_address)
                if bjd_nm is None:
                    return self._create_generate_pnu_result(
                        error=True,
                        msg=f"Failed to extract bjd name from address: {address}"
                    )

                bjd_nm = bjd_nm.strip()
                jibun = main_address.replace(bjd_nm, "")
                if jibun is None: 
                    return self._create_generate_pnu_result(
                        error=True,
                        msg=f"Failed to extract valid jibun from address: {address}"
                    )

                jibun = jibun.strip()
                return self.generate_pnu_from_bjd_nm(bjd_nm=bjd_nm, jibun=jibun)

        except Exception as e:
            return self._create_generate_pnu_result(
                error=True,
                msg=str(e)
            )



import re
import hashlib
from datetime import date
from typing import (
    List,
    Union
)
from dataclasses import dataclass
from vdutils.library import Log


@dataclass
class Vid():


    def __init__(self):
        self.logger = Log('Vid').stream_handler("INFO")
        self.is_pnu_valid_regex = re.compile("^\d{19}$")


    @staticmethod
    def _remove_comma(value: str) -> str:
        return value.replace(",", "")


    @staticmethod
    def _format_float_string(float_str: str) -> str:
        if float_str:
            dot_split = float_str.split(".")
            if len(dot_split) == 1:
                return float_str
            elif len(dot_split) == 2:
                part_int, part_float = dot_split
                float_numbers = list(part_float)
                for i in range(len(part_float) - 1, -1, -1):
                    number = float_numbers[i]
                    if number != "0":
                        break
                    else:
                        float_numbers.pop()
                no_trailing_zero_float_part = "".join(float_numbers)
                if no_trailing_zero_float_part:
                    return ".".join([part_int, no_trailing_zero_float_part])
                else:
                    return part_int
            else:
                return float_str
        else:
            return ""


    def _format_float(
        self, 
        float_num: str
    ) -> str:
        return self._format_float_string(self._remove_comma(str(float_num)))


    def _is_number_valid(
        self, 
        number: Union[float, int, str, None]
    ) -> bool:
        try:
            ff = self._format_float(number)
            if (
                not ff
                or len(ff.split(".")) > 2
                or not re.compile("^\d*$").match(ff.replace(".", ""))
            ):
                return False
            else:
                return True
        except:
            return False


    def _is_pnu_valid(
        self, 
        pnu: str
    ):
        return self.is_pnu_valid_regex.match(pnu)


    @staticmethod
    def _is_contract_date_valid(
        contract_ymd:  Union[str, int, None]
    ) -> bool:
        if type(contract_ymd) != str:
            contract_ymd = str(contract_ymd)
        if len(contract_ymd) == 8 and re.compile("^\d{8}$").match(contract_ymd):
            try:
                d = date.fromisoformat(
                    f"{contract_ymd[:4]}-{contract_ymd[4:6]}-{contract_ymd[6:]}"
                )
                return True
            except:
                return False
        return False


    @staticmethod
    def _is_string_valid(value) -> bool:
        return isinstance(value, str)


    @staticmethod
    def _validate(validator, value, value_nm):
        if value and not validator(value):
            raise Exception(f"invalid {validator}, {value_nm}: {value}")


    @staticmethod
    def _get_hash_raw(data_str: str) -> str:
        return hashlib.sha256(data_str.encode("utf-8")).hexdigest()


    def generate_registration_vid(
        self,
        pnu: str,
        contract_ymd: Union[str, int, None],
        price: Union[float, int, None] = None,
        unit_ar: Union[float, int, str, None] = None,
        lot_ar: Union[float, int, str, None] = None,
        seller: Union[str, None] = None,
        buyer: Union[str, None] = None,
    ) -> List[Union[str, None]]:

        """
        주어진 입력값들을 문자열로 변환하여 SHA-256 해시값을 생성하여 반환\n 
        각 입력값들의 타입정의 및 형식이 명확하지 않아, TypeError, ValueError 를 정의하지 않고 모든 오류에 대해 동일한 예외용 응답만을 반환

        Args:
            pnu (str): Represents the Parcel Numbering Unit, a 19-digit numeric string (10 digits for the administrative district code + 1 digit for the mountain code + 8 digits for the plot number).
            contract_ymd (str): Transaction date in YYYYMMDD format, an 8-digit numeric string.
            price (Union[float, int, None]): Transaction price.
            unit_ar (Union[float, int, str, None]): Transaction unit area.
            lot_ar (Union[float, int, str, None]): Transaction land area.
            seller (Union[str, None]): Seller.
            buyer (Union[str, None]): Buyer.

        Returns: List[Union[str, None]]
        """

        except_response = [
            f"R_{pnu[:10] if self._is_pnu_valid(pnu) else 'pnu10dhead'}_{'hashstring'}_0000",
            None,
            None,
        ]
        try:
            self._validate(self._is_pnu_valid, pnu, "pnu")
            self._validate(self._is_contract_date_valid, contract_ymd, "contract_ymd")
            self._validate(self._is_number_valid, price, "price")
            if not unit_ar == "-":
                self._validate(self._is_number_valid, unit_ar, "unit_ar")
            if not lot_ar == "-":
                self._validate(self._is_number_valid, lot_ar, "lot_ar")
            self._validate(self._is_string_valid, seller, "seller")
            self._validate(self._is_string_valid, buyer, "buyer")

            data_str = "_".join(
                [
                    pnu,
                    contract_ymd,
                    self._format_float(price) if price else "",
                    self._format_float(unit_ar) if price else "",
                    self._format_float(lot_ar) if price else "",
                    seller or "",
                    buyer or "",
                ]
            )
            h = self._get_hash_raw(data_str)

            return [f"R_{pnu[:10]}_{h[:10]}_0000", h, data_str]

        except Exception as e:
            self.logger.info(f"{pnu}, {e}")
            return except_response



import os
import re
import json
import pickle
from typing import (
    List,
    Dict,
    Optional
)
import requests
import pandas as pd
from collections import defaultdict
from dotenv import load_dotenv
from dataclasses import dataclass
from vdutils.resources import resource_filename
from vdutils.library import Log
from vdutils.library.data import (
    ADD_BJD_CHANGED_DICTIONARY,
    CORRECT_ERROR_BJD,
    ADD_BJD,
    DELETE_BJD
)
from vdutils.data import (
    __sep__,
    __index__,
    __encoding__
)


@dataclass
class Bjd():

    def __init__(self):

        load_dotenv()
        try: self.api_key: str = os.environ['BJD_API_KEY']
        except: self.api_key = None
        try: self.api_base_url: str = os.environ['BJD_API_BASE_URL'] # https://www.data.go.kr/iim/api/selectAPIAcountView.do
        except: self.api_base_url = None
        try: self.api_get_url: str = os.environ['BJD_API_SUB_URL'] # https://www.data.go.kr/data/15063424/fileData.do#layer-api-guide API 목록 중 국토교통부_전국 법정동_20230710 GET
        except: self.api_get_url = None

        self.sep = __sep__
        self.index = __index__
        self.encoding = __encoding__
        self.update_dir_path: str = "vdutils/data/update" # 행정안전부 https://www.mois.go.kr/frt/bbs/type001/commonSelectBoardList.do?bbsId=BBSMSTR_000000000052
        self.api_page: int = 0
        self.api_per_page: int = 1024
        self.bjd_api_dictionary: Dict[str, Dict[str, str]] = None
        self.bjd_api_df: pd.DataFrame = None
        self.file_name_bjd = resource_filename(
            "vdutils", 
            "data/bjd.txt"
        )
        self.file_name_bjd_current = resource_filename(
            "vdutils", 
            "data/bjd_current.txt"
        )
        self.file_name_bjd_changed = resource_filename(
            "vdutils", 
            "data/bjd_changed.txt"
        )
        self.file_name_bjd_smallest = resource_filename(
            "vdutils", 
            "data/bjd_smallest.txt"
        )
        self.file_name_bjd_frequency_dictionary = resource_filename(
            "vdutils", 
            "data/bjd_frequency_dictionary.txt"
        )
        self.file_name_multiple_word_sgg_list = resource_filename(
            "vdutils", 
            "data/multiple_word_sgg_list.txt"
        )
        self.logger = Log('Bjd').stream_handler("INFO")
        self.add_bjd_changed_dictionary: Dict[str, str] = ADD_BJD_CHANGED_DICTIONARY
        self.correct_error_bjd: Dict[Dict[str, Optional[str]]] = CORRECT_ERROR_BJD
        self.add_bjd: List[Dict[str, Optional[str]]] = ADD_BJD
        self.delete_bjd: List[str] = DELETE_BJD
        self.multiple_word_sgg_list: List[str] = list()

    @staticmethod
    def _request_api(api_url):
        return requests.get(api_url)

    @staticmethod
    def _convert_json(response):
        return json.loads(response.content)

    def _crawl_api(self) -> Dict[str, Dict[str, str]]:
        res_dic: Dict[str, Dict[str, str]] = dict()
        api_page = self.api_page
        while True:
            api_url = f"{self.api_base_url}{self.api_get_url}?page={api_page}&perPage={self.api_per_page}&serviceKey={self.api_key}"
            print(api_url)
            response = self._request_api(api_url)
            json_datas = self._convert_json(response)
            if json_datas['data']:
                self.logger.info(f"Crawling... {self.api_base_url}{self.api_get_url}?page={api_page}&perPage={self.api_per_page}")
                for data in json_datas['data']:
                    res_dic[f"{str(data['법정동코드'])}"] = {
                        '과거법정동코드': str(data['과거법정동코드']) if data['과거법정동코드'] is not None else None,
                        '리명': str(data['리명']) if data['리명'] is not None else None,
                        '법정동코드': str(data['법정동코드']) if data['법정동코드'] is not None else None,
                        '삭제일자': data['삭제일자'],
                        '생성일자': data['생성일자'],
                        '순위': data['순위'],
                        '시군구명': self._split_sgg_nm(str(data['시군구명'])) if data['시군구명'] is not None else None,
                        '시도명': str(data['시도명']) if data['시도명'] is not None else None,
                        '읍면동명': str(data['읍면동명']) if data['읍면동명'] is not None else None,
                    }
                api_page += 1
            else:
                break
        return res_dic


    def _correct_error(
        self,
        api_dic: Dict[str, Dict[str, str]]
    ):
        for cor_bjd_cd, cor_values in self.correct_error_bjd.items():
            for col_nm, cor_value in cor_values.items():
                api_dic[cor_bjd_cd][col_nm] = cor_value

        for add_bjd_data in self.add_bjd:
            api_dic[add_bjd_data['법정동코드']] = add_bjd_data

        for del_bjd in self.delete_bjd:
            del api_dic[del_bjd]

        return api_dic


    def _update_data(
        self,
        res_dic: Dict[str, Dict[str, str]]
    ) -> Dict[str, Dict[str, str]]:
        if not os.path.exists(self.update_dir_path):
            self.logger(f"Directory does not exist: {self.update_dir_path}")
            return res_dic

        if not os.listdir(self.update_dir_path):
            self.logger(f"No files found in the directory: {self.update_dir_path}")
            return res_dic

        for file_path in sorted(os.listdir(self.update_dir_path)):
            with open(f"{self.update_dir_path}/{file_path}", 'rb') as file:
                update_dic = pickle.load(file)
                res_dic.update(update_dic)
        return res_dic


    def _correct_prev_bjd_cd(
        self,
        sido_nm: str,
        prev_bjd_cd: str, # 법정동코드_변경전
        bjd_cd: str # 법정동코드_변경후
    ):
        if sido_nm == '강원특별자치도':
            return f'42{bjd_cd[2:]}'
        elif bjd_cd in self.add_bjd_changed_dictionary.keys():
            return self.add_bjd_changed_dictionary[bjd_cd]
        else:
            return prev_bjd_cd


    @staticmethod
    def _update_sejongsi(sgg_nm):
        if sgg_nm == '세종시':
            return None
        return sgg_nm


    def _split_sgg_nm(
        self,
        sgg_nm: Optional[str]
    ) -> Optional[str]:
        """
        시군구 행정구역명에서 시와 군 혹은 구가 결합되어있는 행정구역명을 분리하는 기능 \n
        조건에 부합하지 않을 경우 sgg_nm 을 그대로 반환한다

        - 조건 1: 결측값이 아님
        - 조건 2: 문자열수가 4개 이상
        - 조건 3: 마지막 문자열이 '구' 혹은 '군'
        - 조건 4: 첫번째와 마지막 문자열을 제외한 나머지 문자열 중 '시' 가 포함
        """

        sgg_nm = self._update_sejongsi(sgg_nm)

        # TODO 임시: 화성시 신규 구 명칭 강제 등록
        forced_entries = [
            "화성시 만세구",
            "화성시 효행구",
            "화성시 병점구",
            "화성시 동탄구",
        ]
        for entry in forced_entries:
            if entry not in self.multiple_word_sgg_list:
                self.multiple_word_sgg_list.append(entry)

        if sgg_nm is not None \
        and len(sgg_nm) > 3 \
        and sgg_nm[-1] in ['구', '군'] \
        and '시' in sgg_nm[1:-1]:
            result = sgg_nm.split('시', 1)
            result[0] = result[0] + '시'
            return_sgg_nm = ' '.join(result)

            # multiple_word_sgg_list 에 대상 시군구명 추가
            if return_sgg_nm not in self.multiple_word_sgg_list:
                self.multiple_word_sgg_list.append(return_sgg_nm)

            return return_sgg_nm
        else:
            return sgg_nm


    @staticmethod
    def _convert_bjd_nm(
        bjd_nm: Optional[str]
    ) -> str:
        """
        평사리의 한자 표기를 올바른 형태로 변환한다
        bjd_nm is None 일 경우 ''을 반환한다
        """

        if bjd_nm is not None and bjd_nm in ['평(坪)사리', '평(平)사리']:
            # 평사리의 한자 표기 변환 (예: 평(坪)사리 -> 평사리(坪沙), 평(平)사리 -> 평사리(平沙))
            bjd_nm = re.sub(r'평\(坪\)사리', '평사리(坪沙)', bjd_nm)
            bjd_nm = re.sub(r'평\(平\)사리', '평사리(平沙)', bjd_nm)
            return bjd_nm
        return bjd_nm


    @staticmethod
    def _clean_bjd_nm(
        bjd_nm: Optional[str]
    ) -> str:
        """
        행정구역명에서 한글과 숫자를 제외하고 삭제하는 기능
        한자 표기가 포함된 괄호는 유지한다
        bjd_nm is None 일 경우 ''을 반환한다
        """

        if bjd_nm is not None:
            # 한자 표기가 포함된 괄호는 유지하고, 다른 특수문자만 제거
            # 한자 범위: \u4e00-\u9fff (CJK Unified Ideographs)
            bjd_nm = re.sub(r'\([^)]*[^\u4e00-\u9fff)][^)]*\)', '', bjd_nm)
            bjd_nm = re.sub(r'[^ 0-9ㄱ-ㅎ가-힣()\u4e00-\u9fff]+', '', bjd_nm)
            return bjd_nm
        return ''


    def _get_full_bjd_nm(
        self,
        sido_nm: Optional[str],
        sgg_nm: Optional[str],
        emd_nm: Optional[str],
        ri_nm: Optional[str]
    ) -> str:
        """
        행정구역명을 결합하여 전체 법정동명을 생성하는 기능
        """

        full_bjd_nm = f'{sido_nm} {sgg_nm} {emd_nm} {ri_nm}'
        full_bjd_nm = full_bjd_nm.strip()  # 맨 앞과 맨 뒤의 공백 제거
        full_bjd_nm = re.sub(r'\s+', ' ', full_bjd_nm)  # 공백이 여러 칸인 것을 한 칸으로 변경
        return full_bjd_nm


    def _make_dataframe(self, res_dic) -> pd.DataFrame:
        res_df = pd.DataFrame(res_dic).T.sort_values('법정동코드').reset_index().drop(columns='index').replace(0, None).replace('0', None)
        res_df['과거법정동코드'] = res_df[[
                '시도명',
                '과거법정동코드',
                '법정동코드'
            ]].apply(lambda x: self._correct_prev_bjd_cd(*x), axis=1)
        res_df['시도명'] = res_df['시도명'].apply(lambda x: self._clean_bjd_nm(x))
        res_df['시군구명'] = res_df['시군구명'].apply(lambda x: self._clean_bjd_nm(x))
        res_df['읍면동명'] = res_df['읍면동명'].apply(lambda x: self._clean_bjd_nm(x))
        res_df['리명'] = res_df['리명'].apply(lambda x: self._clean_bjd_nm(x))
        res_df['리명'] = res_df['리명'].apply(lambda x: self._convert_bjd_nm(x)) # NOTE 평사리 한자 표기 변환
        res_df['시도명'] = res_df['시도명'].apply(lambda x: x.replace(' ', ''))
        # res_df['시군구명'] = res_df['시군구명'].apply(lambda x: x.replace(' ', ''))
        res_df['읍면동명'] = res_df['읍면동명'].apply(lambda x: x.replace(' ', ''))
        res_df['리명'] = res_df['리명'].apply(lambda x: x.replace(' ', ''))
        res_df['법정동명'] = res_df[[
            '시도명',
            '시군구명',
            '읍면동명',
            '리명'
        ]].apply(lambda x: self._get_full_bjd_nm(*x), axis=1)
        res_df = res_df.reset_index(drop=True)
        return res_df


    def _create_bjd(self):
        """
        국토교통부 전국 법정동 API 수집하여 딕셔너리와 데이터프레임으로 가공하는 기능
        """

        res_dic = self._crawl_api()
        res_dic = self._update_data(res_dic)
        res_dic = self._correct_error(res_dic)
        res_df = self._make_dataframe(res_dic)
        self.bjd_api_dictionary = res_dic
        self.bjd_api_df = res_df
        self.logger.info("Success Created Bjd Dataframe and Dictionary from Public Data API")


    def _save_bjd(self):
        if self.bjd_api_df is None:
            self._create_bjd()
        self.bjd_api_df.to_csv(
            self.file_name_bjd,
            encoding=self.encoding,
            sep=self.sep,
            index=self.index
        )
        self.logger.info("Success Saved Bjd Dataframe To Text File")
        if self.multiple_word_sgg_list is not None:
            with open(self.file_name_multiple_word_sgg_list, 'w', encoding=self.encoding) as f:
                f.writelines('\n'.join(self.multiple_word_sgg_list))
                f.close()
        self.logger.info("Success Saved Multiple Sigungu List To Text File")


    def _do_all_bjd_job(self):
        """
        국토교통부 전국 법정동 API 수집하여 법정동 관련 파일을 모두 생성하는 기능
        - bjd.txt
        - bjd_current.txt
        - bjd_changed.txt
        - bjd_smallest.txt
        - bjd_frequency_dictionary.txt
        """

        self._create_bjd()
        current_bjd = CurrentBjd()
        changed_bjd = ChangedBjd()
        smallest_bjd = SmallestBjd()
        bjd_frequency_dictionary = BjdFrequencyDictionary()

        current_bjd.bjd_api_df = self.bjd_api_df
        current_bjd.bjd_api_dictionary = self.bjd_api_dictionary
        current_bjd._create_current_bjd()
        changed_bjd.bjd_api_df = self.bjd_api_df
        changed_bjd.bjd_api_dictionary = self.bjd_api_dictionary
        changed_bjd._create_changed_bjd()
        
        smallest_bjd.bjd_api_df = self.bjd_api_df
        smallest_bjd.bjd_api_dictionary = self.bjd_api_dictionary
        smallest_bjd.current_bjd_df = current_bjd.current_bjd_df
        smallest_bjd._create_smallest_bjd()
        bjd_frequency_dictionary.bjd_api_df = self.bjd_api_df
        bjd_frequency_dictionary.bjd_api_dictionary = self.bjd_api_dictionary
        bjd_frequency_dictionary.current_bjd_df = current_bjd.current_bjd_df
        bjd_frequency_dictionary._create_bjd_frequency_dictionary()

        self._save_bjd()
        current_bjd._save_current_bjd()
        changed_bjd._save_changed_bjd()
        smallest_bjd._save_smallest_bjd()
        bjd_frequency_dictionary._save_bjd_frequency_dictionary()


@dataclass
class CurrentBjd(Bjd):


    def __init__(self):
        super().__init__()
        self.current_bjd_df: pd.DataFrame = None


    def _create_current_bjd(self):
        """
        국토교통부 전국 법정동 API 현재 존재하는 법정동만 데이터프레임으로 가공하는 기능
        """

        if self.bjd_api_df is None:
            self._create_bjd()
        self.current_bjd_df = self.bjd_api_df.loc[self.bjd_api_df['삭제일자'].isnull()]
        self.logger.info("Success Created Current Bjd Dataframe")


    def _save_current_bjd(self):
        if self.current_bjd_df is None:
            self._create_current_bjd()
        self.current_bjd_df.to_csv(
            self.file_name_bjd_current,
            encoding=self.encoding,
            sep=self.sep,
            index=self.index
        )
        self.logger.info("Success Saved Current Bjd Dataframe To Text File")


@dataclass
class ChangedBjd(Bjd):


    def __init__(self):
        super().__init__()
        self.changed_bjd_df: pd.DataFrame = None


    def _get_prev_bjd_nm(
        self,
        prev_bjd_cd: Optional[str] # 법정동코드_변경전
    ) -> str:
        if prev_bjd_cd is None:
            return None
        if prev_bjd_cd not in self.bjd_api_dictionary.keys():
            return None
        try:
            data = self.bjd_api_dictionary[prev_bjd_cd]
            sido_nm = self._clean_bjd_nm(data['시도명']).replace(' ', '')
            sgg_nm = self._clean_bjd_nm(data['시군구명'])
            emd_nm = self._clean_bjd_nm(data['읍면동명']).replace(' ', '')
            ri_nm = self._clean_bjd_nm(data['리명']).replace(' ', '')

            return self._get_full_bjd_nm(
                sido_nm,
                sgg_nm,
                emd_nm,
                ri_nm
            )
        except:
            return None


    def _get_prev_value(
        self,
        prev_bjd_cd: Optional[str], # 법정동코드_변경전
        value_nm: str
    ):
        if prev_bjd_cd is None:
            return None
        if prev_bjd_cd not in self.bjd_api_dictionary.keys():
            return None
        try:
            return self.bjd_api_dictionary[prev_bjd_cd][value_nm]
        except:
            return None


    @staticmethod
    def _find_diff(
        bjd_cd_curr: Optional[str], # 법정동코드_변경전
        bjd_cd_prev: Optional[str], # 법정동코드_변경후
        bjd_nm_curr: Optional[str], # 법정동명_변경전
        bjd_nm_prev: Optional[str] # 법정동명_변경후
    ) -> str:
        bjd_cd_changed = ''
        bjd_nm_changed = ''
        if bjd_cd_prev and bjd_cd_curr:
            if bjd_cd_prev != bjd_cd_curr:
                bjd_cd_changed = f'{bjd_cd_prev} > {bjd_cd_curr}'
        if bjd_nm_prev and bjd_nm_curr:
            changed_list_prev = list()
            changed_list_curr = list()
            for w1, w2 in zip(bjd_nm_prev.split(), bjd_nm_curr.split()):
                if w1 != w2:
                    changed_list_prev.append(w1)
                    changed_list_curr.append(w2)
            bjd_nm_changed = f"{' '.join(changed_list_prev)} > {' '.join(changed_list_curr)}"
        return f'[법정동코드 변경내역] {bjd_cd_changed} | [법정동명 변경내역]: {bjd_nm_changed}'


    def _create_changed_bjd(self):
        if self.bjd_api_df is None:
            self._create_bjd()
        self.changed_bjd_df = self.bjd_api_df.copy()
        self.changed_bjd_df = self.changed_bjd_df.rename(columns={
            '법정동코드': '법정동코드_변경후',
            '법정동명': '법정동명_변경후',
            '생성일자': '생성일자_변경후',
            '삭제일자': '삭제일자_변경후',
            '과거법정동코드': '법정동코드_변경전',
        })
        self.changed_bjd_df = self.changed_bjd_df[[
            '법정동코드_변경후',
            '법정동명_변경후',
            '생성일자_변경후',
            '삭제일자_변경후',
            '법정동코드_변경전'
        ]]
        self.changed_bjd_df = self.changed_bjd_df.loc[
            (self.changed_bjd_df['생성일자_변경후'] != self.changed_bjd_df['삭제일자_변경후']) &
            ~(self.changed_bjd_df['생성일자_변경후'] > self.changed_bjd_df['삭제일자_변경후'])
        ]
        self.changed_bjd_df = self.changed_bjd_df.loc[
            (self.changed_bjd_df['법정동코드_변경전'].isnull()==False) &
            (self.changed_bjd_df['법정동코드_변경전'].str.len() == 10)
        ].sort_values('생성일자_변경후')
        self.changed_bjd_df['법정동명_변경전'] = self.changed_bjd_df['법정동코드_변경전'].apply(lambda x: self._get_prev_bjd_nm(x))
        self.changed_bjd_df['생성일자_변경전'] = self.changed_bjd_df['법정동코드_변경전'].apply(lambda x: self._get_prev_value(x, '생성일자'))
        self.changed_bjd_df['삭제일자_변경전'] = self.changed_bjd_df['법정동코드_변경전'].apply(lambda x: self._get_prev_value(x, '삭제일자'))

        self.changed_bjd_df['변경내역'] = self.changed_bjd_df[[
            '법정동코드_변경후',
            '법정동코드_변경전', 
            '법정동명_변경후',
            '법정동명_변경전'
        ]].apply(lambda x: self._find_diff(*x), axis=1)
        self.changed_bjd_df = self.changed_bjd_df.sort_values(['생성일자_변경후'], na_position='first')
        self.logger.info("Success Created Changed Bjd Dataframe")


    def _save_changed_bjd(self):
        if self.changed_bjd_df is None:
            self._create_changed_bjd()
        self.changed_bjd_df.to_csv(
            self.file_name_bjd_changed,
            encoding=self.encoding,
            sep=self.sep,
            index=self.index
        )
        self.logger.info("Success Saved Changed Bjd Dataframe To Text File")


@dataclass
class SmallestBjd(CurrentBjd):


    def __init__(self):
        super().__init__()
        self.smallest_bjd_list: List[str] = None


    def _create_smallest_bjd(self):
        if self.current_bjd_df is None:
            self._create_current_bjd()
        smallest_bjd_set = set()
        for bjd in self.current_bjd_df['법정동명']:
            if bjd[-1] in ['가', '동', '로', '리']:
                smallest_bjd_set.add(bjd.split(' ')[-1])

        self.smallest_bjd_list = sorted(list(smallest_bjd_set))
        self.logger.info("Success Created Smallest Bjd Name List")


    def _save_smallest_bjd(self):
        if self.smallest_bjd_list is None:
            self._create_smallest_bjd()
        with open(self.file_name_bjd_smallest, 'w', encoding=self.encoding) as f:
            f.writelines('\n'.join(self.smallest_bjd_list))
            f.close()
        self.logger.info("Success Saved Changed Smallest Bjd Name List To Text File")


@dataclass
class BjdFrequencyDictionary(CurrentBjd):


    def __init__(self):
        super().__init__()
        self.bjd_frequency_dictionary: Dict[str, int] = None


    def _create_bjd_frequency_dictionary(self):
        if self.current_bjd_df is None:
            self._create_current_bjd()
        self.bjd_frequency_dictionary = defaultdict(int)
        for bjd in self.current_bjd_df['법정동명']:
            bjd_words = bjd.split(' ')
            for word in bjd_words:
                self.bjd_frequency_dictionary[word] += 1

        self.logger.info("Success Created Bjd Frequency Dictionary")


    def _save_bjd_frequency_dictionary(self):
        if self.bjd_frequency_dictionary is None:
            self._create_bjd_frequency_dictionary()
        bjd_frequency_list = list((key, value) for key, value in self.bjd_frequency_dictionary.items())
        with open(self.file_name_bjd_frequency_dictionary, 'w', encoding=self.encoding) as f:
            vstr = ''
            sep = ','
            for line in bjd_frequency_list:
                for s in line:
                    vstr = vstr + str(s) + sep
                vstr = vstr.rstrip(sep)  # 마지막에도 추가되는  sep을 삭제 
                vstr = vstr + '\n'
            f.writelines(vstr)  # 한 라인씩 저장 
            f.close()

        self.logger.info("Success Saved Bjd Frequency Dictionary To Text File")



doc = """
vos-data-utils - a utility library for ValueOfSpace data work with Python
==========================================================================

Main Features
-------------
Correction function for dates related to building permit, commencement, and completion in the Building Permit Information
Support for current legal district data and change history
Correction function to align with the current legal district by reflecting changes in legal district data
Conversion function to transform address strings into Parcel Number (PNU)
Generation function for unique transaction case IDs in ValueofSpace
"""
version = "1.0.15" 
author = "ValueOfSpace"
description = "description"
license = "MIT License"
python_version = "3.8, 3.9, 3.10, 3.11, 3.12"

def __version__():
    return version

def __author__():
    return author

def __description__():
    return description

def __license__():
    return license

def __doc__():
    return doc



import os
import sys
import unittest
import logging
from typing import Pattern
sys.path.append(os.getcwd())
from vdutils.vid import Vid


class TestClass(unittest.TestCase):

    @classmethod
    def setUp(cls):
        "Hook method for setting fixture before running tests in the class"
        cls.driver = 'test'
        cls.instance = Vid()


    @classmethod
    def tearDown(cls):
        "Hook method for deconstructing the class fixture after running all tests in the class"


    def test_class_initialization_type(self):
        """
        클래스 인스턴스 초기 생성자 타입 테스트 메소드

        - 클래스 인스턴스 초기 생성자 객체의 값의 타입이 지정된 타입과 일치하는지 확인하는 테스트
            - __init__.logger 객체 테스트
            - __init__.is_pnu_valid_regex 객체 테스트
        """

        self.assertIsInstance(self.instance.logger, logging.Logger)
        self.assertIsInstance(self.instance.is_pnu_valid_regex, Pattern) # default False


    def test_class_initialization_not_empty(self):
        """
        클래스 인스턴스 초기 생성자 객체 테스트 메소드

        - 클래스 인스턴스 초기 생성자 객체의 값이 None이 아닌지(IsNotNone) 확인하는 테스트
            - __init__.logger 객체 테스트
            - __init__.is_pnu_valid_regex 객체 테스트

        - 클래스 인스턴스 초기 생성자 객체의 Length가 0이 아닌지 확인하는 테스트
            - __init__.logger 객체 테스트 (logging.Logger pass)
            - __init__.is_pnu_valid_regex 객체 테스트 (pattern pass)
        """

        self.assertIsNotNone(self.instance.logger)
        self.assertIsNotNone(self.instance.is_pnu_valid_regex)

        # self.assertNotEqual(len(self.instance.logger), 0)
        # self.assertNotEqual(len(self.instance.is_pnu_valid_regex), 0)


    def test_generate_registration_vid(self):
        """
        generate_registration_vid 함수 테스트 메소드

        - generate_registration_vid 함수 리턴값 테스트(올바른 입력값 제공)
            - generate_registration_vid 함수 리턴값 타입 테스트
            - generate_registration_vid 함수 리턴값 None 이 아닌지 테스트
            - generate_registration_vid 함수 리턴값 리스트 첫번째 객체값 타입 테스트
            - generate_registration_vid 함수 리턴값 리스트 첫번째 객체값 None 이 아닌지 테스트
            - generate_registration_vid 함수 리턴값 리스트 두번째 객체값 타입 테스트
            - generate_registration_vid 함수 리턴값 리스트 두번째 객체값 None 이 아닌지 테스트
            - generate_registration_vid 함수 리턴값 리스트 세번째 객체값 타입 테스트
            - generate_registration_vid 함수 리턴값 리스트 세번째 객체값 None 이 아닌지 테스트
            - generate_registration_vid 함수 리턴값 일치 테스트

        - generate_registration_vid 함수 리턴값 테스트(올바르지 않은 입력값 제공)
            - generate_registration_vid 함수 리턴값 타입 테스트
            - generate_registration_vid 함수 리턴값 None 이 아닌지 테스트
            - generate_registration_vid 함수 리턴값 리스트 첫번째 객체값 타입 테스트
            - generate_registration_vid 함수 리턴값 리스트 첫번째 객체값 None 이 아닌지 테스트
            - generate_registration_vid 함수 리턴값 리스트 두번째 객체값 타입 테스트 (None pass)
            - generate_registration_vid 함수 리턴값 리스트 두번째 객체값 None 이 맞는지 테스트
            - generate_registration_vid 함수 리턴값 리스트 세번째 객체값 타입 테스트 (None pass)
            - generate_registration_vid 함수 리턴값 리스트 세번째 객체값 None 이 맞는지 테스트
            - generate_registration_vid 함수 리턴값 일치 테스트
        """

        res = self.instance.generate_registration_vid(
            pnu='1234567890123456789',
            contract_ymd='20240101',
            price=100,
            unit_ar=100.1,
            lot_ar=100.1,
            seller='seller',
            buyer='buyer',
        )
        self.assertIsInstance(res, list)
        self.assertIsNotNone(res)
        self.assertIsInstance(res[0], str)
        self.assertIsNotNone(res[0])
        self.assertIsInstance(res[1], str)
        self.assertIsNotNone(res[1])
        self.assertIsInstance(res[2], str)
        self.assertIsNotNone(res[2])
        self.assertEqual(res, [
            'R_1234567890_581e28e445_0000',
            '581e28e4457103048784572e858c5ed1297cf8747169a8454e98a19233036f55',
            '1234567890123456789_20240101_100_100.1_100.1_seller_buyer'
        ])

        res = self.instance.generate_registration_vid(
            pnu='1234567890123456789',
            contract_ymd='20240101',
            price='100',
            unit_ar='100.1',
            lot_ar='100.1',
            seller='seller',
            buyer='buyer',
        )
        self.assertIsInstance(res, list)
        self.assertIsNotNone(res)
        self.assertIsInstance(res[0], str)
        self.assertIsNotNone(res[0])
        self.assertIsInstance(res[1], str)
        self.assertIsNotNone(res[1])
        self.assertIsInstance(res[2], str)
        self.assertIsNotNone(res[2])
        self.assertEqual(res, [
            'R_1234567890_581e28e445_0000',
            '581e28e4457103048784572e858c5ed1297cf8747169a8454e98a19233036f55',
            '1234567890123456789_20240101_100_100.1_100.1_seller_buyer'
        ])

        res = self.instance.generate_registration_vid(
            pnu='12345678901234567890', # 19자리 형식에 맞지 않는 잘못된 입력값 제공
            contract_ymd='20240101',
            price=100,
            unit_ar=100.1,
            lot_ar=100.1,
            seller='seller',
            buyer='buyer',
        )
        self.assertIsInstance(res, list)
        self.assertIsNotNone(res)
        self.assertIsInstance(res[0], str)
        self.assertIsNotNone(res[0])
        # self.assertIsInstance(res[1], None) # (None pass)
        self.assertIsNone(res[1])
        # self.assertIsInstance(res[2], None) # (None pass)
        self.assertIsNone(res[2])
        self.assertEqual(res, [
            'R_pnu10dhead_hashstring_0000',
            None,
            None
        ])

        res = self.instance.generate_registration_vid(
            pnu='1234567890123456789',
            contract_ymd='2024-01-01',  # YYYYMMDD 형식에 맞지 않는 잘못된 입력값 제공
            price=100,
            unit_ar=100.1,
            lot_ar=100.1,
            seller='seller',
            buyer='buyer',
        )
        self.assertIsInstance(res, list)
        self.assertIsNotNone(res)
        self.assertIsInstance(res[0], str)
        self.assertIsNotNone(res[0])
        # self.assertIsInstance(res[1], None) # (None pass)
        self.assertIsNone(res[1])
        # self.assertIsInstance(res[2], None) # (None pass)
        self.assertIsNone(res[2])
        self.assertEqual(res, [
            'R_1234567890_hashstring_0000',
            None,
            None
        ])

        res = self.instance.generate_registration_vid(
            pnu='1234567890123456789',
            contract_ymd='20240101',
            price=-100, # 음수 잘못된 입력값 제공
            unit_ar=100.1,
            lot_ar=100.1,
            seller='seller',
            buyer='buyer',
        )
        self.assertIsInstance(res, list)
        self.assertIsNotNone(res)
        self.assertIsInstance(res[0], str)
        self.assertIsNotNone(res[0])
        # self.assertIsInstance(res[1], None) # (None pass)
        self.assertIsNone(res[1])
        # self.assertIsInstance(res[2], None) # (None pass)
        self.assertIsNone(res[2])
        self.assertEqual(res, [
            'R_1234567890_hashstring_0000',
            None,
            None
        ])

        res = self.instance.generate_registration_vid(
            pnu='1234567890123456789',
            contract_ymd='20240101',
            price=100,
            unit_ar='a', # 숫자 혹은 숫자문자열이 아닌 문자열 입력값 제공
            lot_ar=100.1,
            seller='seller',
            buyer='buyer',
        )
        self.assertIsInstance(res, list)
        self.assertIsNotNone(res)
        self.assertIsInstance(res[0], str)
        self.assertIsNotNone(res[0])
        # self.assertIsInstance(res[1], None) # (None pass)
        self.assertIsNone(res[1])
        # self.assertIsInstance(res[2], None) # (None pass)
        self.assertIsNone(res[2])
        self.assertEqual(res, [
            'R_1234567890_hashstring_0000',
            None,
            None
        ])

        res = self.instance.generate_registration_vid(
            pnu='1234567890123456789',
            contract_ymd='20240101',
            price=100,
            unit_ar=100.1,
            lot_ar=100.1,
            seller=1, # 문자열 아닌 숫자 입력값 제공
            buyer='buyer',
        )
        self.assertIsInstance(res, list)
        self.assertIsNotNone(res)
        self.assertIsInstance(res[0], str)
        self.assertIsNotNone(res[0])
        # self.assertIsInstance(res[1], None) # (None pass)
        self.assertIsNone(res[1])
        # self.assertIsInstance(res[2], None) # (None pass)
        self.assertIsNone(res[2])
        self.assertEqual(res, [
            'R_1234567890_hashstring_0000',
            None,
            None
        ])


if __name__ == "__main__":
    unittest.main()


import setuptools
from vdutils import (
    author,
    version,
    description,
    license
)


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="vos-data-utils",
    version=version,
    author=author,
    author_email="dev@valueofspace.com",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={"vdutils": [
        "data/bjd/*/*.txt", 
        "data/date/*.txt"
    ]},
    classifiers=[
        f"License :: OSI Approved :: {license}",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    python_requires='>=3.8',
    install_requires=[
        "symspellpy",
        "pandas",
        "requests",
        "python-dotenv"
    ],
    entry_points={
        'console_scripts': [
            'shortcut1 = package.module:func',
        ],
        'gui_scripts': [
            'shortcut2 = package.module:func',
        ]
    }
)



import os
from typing import (
    List
)
from vdutils.resources import resource_filename


__sep__ = '\t'
__encoding__: str = 'utf-8'
__index__: bool = False

def _get_folder_names(
    library_name: str,
    base_folder_path: str
) -> List[str]:
    resource_path = resource_filename(library_name, base_folder_path)
    return [f.name for f in os.scandir(resource_path) if f.is_dir()]


from datetime import datetime
from typing import (
    Optional,
    List
)
from symspellpy import (
    SymSpell,
    Verbosity
)
from dataclasses import dataclass
from vdutils.resources import resource_filename


@dataclass
class CorDate():


    def __init__(self):
        self.sym_spell = SymSpell()
        self.dictionary_path = resource_filename(
            "vdutils", 
            "data/date/date_dictionary.txt"
        )
        self.this_year: int = datetime.now().year
        self.this_year_two_length: int = int(str(self.this_year)[2:])
        self.max_edit_distance: int = 2


    def load_date_dictionary(self) -> bool:

        """
        라이브러리 배포 폴더에 있는 date_dictionary.txt 로드
        Symspellpy를 이용하는 look_up_array, look_up_one 전에 수행필요
        """

        try:
            self.sym_spell.load_dictionary(self.dictionary_path, 0, 1, separator="$")
            return True
        except:
            return False


    def get_correct_array(
        self, 
        date: str
    ) -> List[str]:

        """
        입력된 문자열을 이용해 날짜 생성 규칙에 따라 현재 날짜까지 생성 가능한 모든 날짜를 리스트로 생성함
        날짜 생성 규칙이란 연,월,일의 범위를 이용하는것으로 연도는 올해연도까지, 월은 1부터 12월까지, 일은 월별로 지정된 일까지
        일반적으로 연도는 4자리, 월, 일은 2자리로 표기하지만 자리수 범위는 각 [0:4],[0:2],[0:2] 차지할 수 있음

        Args:
            date (str): The input date string to search for.

        Raises:
            TypeError: If the 'date' object is not of type string.
            ValueError: If the 'date' object does not consist of digits only.

        Returns:
            List[str]: The closest value in the string of dates, or None if not found.
        """

        if not isinstance(date, str):
            raise TypeError("type of object must be string")

        if not date.isdigit():
            raise ValueError("object('date') should be a string consisting of numbers")

        dates = _check_correct_ymd(date, self.this_year)

        if _check_two_length_year(dates):
            dates = _get_four_length_year(dates, self.this_year_two_length)
        return dates


    def get_correct_one(
        self, 
        date: str
    ) -> Optional[str]:

        """
        입력된 문자열을 이용해 날짜 생성 규칙에 따라 현재 날짜까지 생성 가능한 모든 날짜 리스트중 가장 최신날짜를 출력
        날짜 생성 규칙이란 연,월,일의 범위를 이용하는것으로 연도는 올해연도까지, 월은 1부터 12월까지, 일은 월별로 지정된 일까지
        일반적으로 연도는 4자리, 월, 일은 2자리로 표기하지만 자리수 범위는 각 [0:4],[0:2],[0:2] 차지할 수 있음

        Args:
            date (str): The input date string to search for.

        Raises:
            TypeError: If the 'date' object is not of type string.
            ValueError: If the 'date' object does not consist of digits only.

        Returns:
            Optional[str]: The closest value in the string of dates, or None if not found.
        """

        if not isinstance(date, str):
            raise TypeError("type of object must be string")

        if not date.isdigit():
            raise ValueError("object('date') should be a string consisting of numbers")

        return _get_correct_date_from_dates(
            _check_correct_ymd(
                date,
                self.this_year
            ),
            self.this_year_two_length
        )


    def look_up_array(
        self,
        date : str,
        max_edit_distance : Optional[int] = None,
    ) -> List[str]:

        """
        연월일 문자열에 Symspellpy로 max_distance=2로 날짜 리스트 출력

        Args:
            date (str): The date string in the format 'yyyymmdd'.
            max_edit_distance (Optional[int]): The maximum edit distance for spell correction. If not provided, defaults to self.max_edit_distance.

        Raises:
            TypeError: If the 'date' object is not of type string.
            ValueError: If the 'date' object does not consist of digits only.

        Returns:
            List[str]: The closest date string from the date list, or None if not found.
        """

        if not isinstance(date, str):
            raise TypeError("type of object('date') must be string")

        if not date.isdigit():
            raise ValueError("object('date') should be a string consisting of numbers")

        if max_edit_distance is None:
            max_edit_distance = self.max_edit_distance

        if max_edit_distance > self.max_edit_distance:
            raise ValueError("distance too large")

        suggestions = self.sym_spell.lookup(
            date,
            Verbosity.ALL,
            max_edit_distance=max_edit_distance
        )
        return suggestions


    def look_up_array_clean(
        self,
        date : str,
        max_edit_distance : Optional[int] = None,
    ) -> List[str]:

        """
        look_up_array 결과값에서 term 리스트출력

        Args:
            date (str): The date string in the format 'yyyymmdd'.
            max_edit_distance (Optional[int]): The maximum edit distance for spell correction. If not provided, defaults to self.max_edit_distance.

        Raises:
            TypeError: If the 'date' object is not of type string.
            ValueError: If the 'date' object does not consist of digits only.

        Returns:
            List[str]: The closest date string from the date list, or None if not found.
        """

        suggestions = self.look_up_array(
            date=date,
            max_edit_distance=max_edit_distance
        )
        if suggestions \
        and len(suggestions):
            suggestions = list(_.term for _ in suggestions)
            return suggestions
        else:
            return suggestions


    def look_up_one(
        self,
        date : str,
        max_edit_distance : Optional[int] = None,
    ) -> Optional[str]:

        """
        연월일 문자열에 Symspellpy로 max_distance=2로 날짜 리스트 중 가장 거리, 빈도 가까운 값 출력

        Args:
            date (str): The date string in the format 'yyyymmdd'.
            max_edit_distance (Optional[int]): The maximum edit distance for spell correction. If not provided, defaults to self.max_edit_distance.

        Raises:
            TypeError: If the 'date' object is not of type string.
            ValueError: If the 'date' object does not consist of digits only.

        Returns:
            Optional[str]: The closest date string from the date list, or None if not found.
        """

        if not isinstance(date, str):
            raise TypeError("type of object('date') must be string")

        if not date.isdigit():
            raise ValueError("object('date') should be a string consisting of numbers")

        if max_edit_distance is None:
            max_edit_distance = self.max_edit_distance

        if max_edit_distance > self.max_edit_distance:
            raise ValueError("distance too large")

        suggestions = self.sym_spell.lookup(
            date,
            Verbosity.ALL,
            max_edit_distance=max_edit_distance
        )
        if len(suggestions):
            suggestion = suggestions[0]
            return suggestion
        else:
            return None


    def look_up_one_clean(
        self,
        date : str,
        max_edit_distance : Optional[int] = None,
    ) -> Optional[str]:

        """
        look_up_one 결과값에서 term 출력

        Args:
            date (str): The date string in the format 'yyyymmdd'.
            max_edit_distance (Optional[int]): The maximum edit distance for spell correction. If not provided, defaults to self.max_edit_distance.

        Raises:
            TypeError: If the 'date' object is not of type string.
            ValueError: If the 'date' object does not consist of digits only.

        Returns:
            Optional[str]: The closest date string from the date list, or None if not found.
        """

        suggestion = self.look_up_one(
            date=date,
            max_edit_distance=max_edit_distance
        )
        if suggestion:
            return suggestion.term
        else:
            return suggestion


def _convert_type_year(y: str) -> int:

    if y == '':
        return 0
    else:
        return int(y)


def _convert_type_month_day(md: str) -> int:

    if md in ['', '0', '00']:
        return 1
    else:
        return int(md)


def _check_correct_ymd(
    date: str, 
    this_year: int
) -> List[str]:

    candidate_dates = set()
    for idx_day in range(3):
        for idx_month in range(3):
            try:
                d, e_date = date[len(date)-idx_day:], date[:len(date)-idx_day]
                m, y = e_date[len(e_date)-idx_month:], e_date[:len(e_date)-idx_month]
                d = _convert_type_month_day(d)
                m = _convert_type_month_day(m)
                y = _convert_type_year(y)
                candidate_date = datetime(y, m, d)
                if (y > 0) and (y < (this_year +1)) and candidate_date:
                    candidate_date = candidate_date.strftime("%Y%m%d").zfill(8)
                    candidate_dates.add(candidate_date)
            except:
                pass
    return list(candidate_dates)


def _check_two_length_year(
    dates: List[str]
) -> bool:

    if dates and all(1000 > int(ymd[:4]) for ymd in dates):
        return True
    else:
        return False


def _change_four_length_year(
    year: str,
    this_year_two_length: int
) -> str:

    if year[0] == '0':
        if year[1] == '0':
            if year[2] == '0': # 00090101처럼 000으로 시작되는 경우
                pass
            else:
                if int(year) in list(range(0, this_year_two_length +1)):
                    year = '20' + year[2:]
                else:
                    year = '19' + year[2:]
        else:
            year = '1' + year[1:]
    return year


def _get_four_length_year(
    dates: List[str], 
    this_year_two_length
) -> List[str]:

    new_candidate_dates = list()
    for date in dates:
        date = _change_four_length_year(date[:4], this_year_two_length) + date[4:]
        new_candidate_dates.append(date)
    return new_candidate_dates


def _get_one_date(
    dates: List[str]
) -> Optional[str]:

    # 가장 최신값
    try:
        if dates and len(dates) > 0:
            return sorted(dates, reverse=True)[0]
        else:
            return None
    except:
        return None


def _get_correct_date_from_dates(
    dates: List[str],
    this_year_two_length: int
) -> Optional[str]:

    if _check_two_length_year(dates):
        dates = _get_four_length_year(dates, this_year_two_length)
    return _get_one_date(dates)


def get_correct_array( 
    date: str
) -> List[str]:

    """
    입력된 문자열을 이용해 날짜 생성 규칙에 따라 현재 날짜까지 생성 가능한 모든 날짜를 리스트로 생성함
    날짜 생성 규칙이란 연,월,일의 범위를 이용하는것으로 연도는 올해연도까지, 월은 1부터 12월까지, 일은 월별로 지정된 일까지
    일반적으로 연도는 4자리, 월, 일은 2자리로 표기하지만 자리수 범위는 각 [0:4],[0:2],[0:2] 차지할 수 있음

    Args:
        date (str): The input date string to search for.

    Raises:
        TypeError: If the 'date' object is not of type string.
        ValueError: If the 'date' object does not consist of digits only.

    Returns:
        List[str]: The closest value in the list of dates, or None if not found.
    """

    if not isinstance(date, str):
        raise TypeError("type of object must be string")

    if not date.isdigit():
        raise ValueError("object('date') should be a string consisting of numbers")

    this_year = datetime.now().year
    this_year_two_length = int(str(this_year)[2:])
    dates = _check_correct_ymd(date, this_year)

    if _check_two_length_year(dates):
        dates = _get_four_length_year(dates, this_year_two_length)
    return dates


def get_correct_one(
    date: str
) -> Optional[str]:

    """
    입력된 문자열을 이용해 날짜 생성 규칙에 따라 현재 날짜까지 생성 가능한 모든 날짜 리스트중 가장 최신날짜를 출력
    날짜 생성 규칙이란 연,월,일의 범위를 이용하는것으로 연도는 올해연도까지, 월은 1부터 12월까지, 일은 월별로 지정된 일까지
    일반적으로 연도는 4자리, 월, 일은 2자리로 표기하지만 자리수 범위는 각 [0:4],[0:2],[0:2] 차지할 수 있음

    Args:
        date (str): The input date string to search for.

    Raises:
        TypeError: If the 'date' object is not of type string.
        ValueError: If the 'date' object does not consist of digits only.

    Returns:
        Optional[str]: The closest value in the string of dates, or None if not found.
    """

    if not isinstance(date, str):
        raise TypeError("type of object must be string")

    if not date.isdigit():
        raise ValueError("object('date') should be a string consisting of numbers")

    this_year = datetime.now().year
    this_year_two_length = int(str(this_year)[2:])
    return _get_correct_date_from_dates(
        _check_correct_ymd(
            date,
            this_year
        ),
        this_year_two_length
    )
